.statistics-container {
    font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
    padding: 20px;
    background: var(--bg-secondary);
    border-radius: 8px;
    color: var(--text-primary);
    transition: all 0.3s ease;
}

.statistics-header {
    margin-bottom: 24px;
    position: relative;
}

.statistics-title {
    margin: 0 0 8px 0;
    font-size: 24px;
    font-weight: 600;
    color: var(--text-primary);
}

.statistics-subtitle {
    margin: 0;
    font-size: 14px;
    color: var(--text-secondary);
}

.theme-toggle {
    position: absolute;
    top: 0;
    right: 0;
    padding: 8px 12px;
    background: var(--bg-primary);
    border: 2px solid var(--border-color);
    border-radius: 6px;
    color: var(--text-primary);
    cursor: pointer;
    display: flex;
    align-items: center;
    gap: 6px;
    font-size: 13px;
    font-weight: 500;
    transition: all 0.2s;
    box-shadow: 0 1px 3px rgba(0, 0, 0, 0.1);
}

.theme-toggle:hover {
    background: var(--hover-bg);
    border-color: var(--text-secondary);
    box-shadow: 0 2px 4px rgba(0, 0, 0, 0.15);
}

.theme-toggle:active {
    transform: scale(0.98);
}

.statistics-grid {
    display: grid;
    grid-template-columns: repeat(auto-fit, minmax(280px, 1fr));
    gap: 16px;
    margin-bottom: 20px;
}

.stat-card {
    background: var(--bg-primary);
    padding: 16px;
    border-radius: 6px;
    box-shadow: 0 1px 3px rgba(0, 0, 0, 0.1);
    border: 1px solid var(--border-light);
    transition: all 0.3s ease;
}

.stat-label {
    margin: 0 0 12px 0;
    font-size: 14px;
    color: var(--text-secondary);
    text-transform: uppercase;
    letter-spacing: 0.5px;
    font-weight: 600;
}

.stat-value {
    display: flex;
    align-items: baseline;
    margin-bottom: 12px;
}

.stat-current {
    font-size: 28px;
    font-weight: bold;
    color: var(--text-primary);
}

.stat-limit {
    color: var(--text-muted);
    margin-left: 8px;
    font-size: 14px;
}

.stat-progress-bg {
    background: var(--border-light);
    border-radius: 4px;
    height: 8px;
    overflow: hidden;
    margin-bottom: 8px;
}

.stat-progress-bar {
    height: 100%;
    transition: width 0.3s ease, background 0.3s ease;
    background: #10b981;
}

.stat-usage {
    font-size: 12px;
    font-weight: 500;
    color: #10b981;
    transition: color 0.3s ease;
}

.statistics-additional {
    background: var(--bg-primary);
    padding: 16px;
    border-radius: 6px;
    box-shadow: 0 1px 3px rgba(0, 0, 0, 0.1);
    border: 1px solid var(--border-light);
    transition: all 0.3s ease;
}

.additional-title {
    margin: 0 0 12px 0;
    font-size: 14px;
    color: var(--text-secondary);
    text-transform: uppercase;
    letter-spacing: 0.5px;
    font-weight: 600;
}

.additional-content {
    display: grid;
    grid-template-columns: 1fr 1fr;
    gap: 12px;
}

.component-item {
    color: var(--text-primary);
    transition: color 0.3s ease;
}

.component-item strong {
    color: var(--text-primary);
}

.component-value {
    color: #0066cc;
    font-weight: 600;
    margin-left: 8px;
    transition: color 0.3s ease;
}

.dark-mode .component-value {
    color: #60a5fa;
}